@@ -3,14 +3,18 @@
  *
  * All rights reserved.
  *
- * This file is part of ObjFW. It may be distributed under the terms of the
- * Q Public License 1.0, which can be found in the file LICENSE.QPL included in
- * the packaging of this file.
+ * This program is free software: you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License version 3.0 only,
+ * as published by the Free Software Foundation.
  *
- * Alternatively, it may be distributed under the terms of the GNU General
- * Public License, either version 2 or 3, which can be found in the file
- * LICENSE.GPLv2 or LICENSE.GPLv3 respectively included in the packaging of this
- * file.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
+ * version 3.0 for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * version 3.0 along with this program. If not, see
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -22,6 +26,7 @@
 #import "OFFile.h"
 #import "OFFileManager.h"
 #import "OFIRI.h"
+#import "OFIRIHandler.h"
 #import "OFLocale.h"
 #import "OFOptionsParser.h"
 #import "OFSandbox.h"
@@ -32,8 +37,10 @@
 #import "LHAArchive.h"
 #import "TarArchive.h"
 #import "ZIPArchive.h"
+#import "ZooArchive.h"
 
 #import "OFCreateDirectoryFailedException.h"
+#import "OFGetItemAttributesFailedException.h"
 #import "OFInvalidArgumentException.h"
 #import "OFInvalidFormatException.h"
 #import "OFNotImplementedException.h"
@@ -57,24 +64,32 @@ help(OFStream *stream, bool full, int status)
 		[stream writeString: @"\n"];
 		[stream writeLine: OF_LOCALIZED(@"full_usage",
 		    @"Options:\n"
-		    @"    -a  --append      Append to archive\n"
-		    @"    -c  --create      Create archive\n"
-		    @"    -C  --directory   Extract into the specified "
+		    @"    -a  --append            Append to archive\n"
+		    @"        --archive-comment=  Archive comment to use when "
+		    @"creating or appending\n"
+		    @"    -c  --create            Create archive\n"
+		    @"    -C  --directory=        Extract into the specified "
 		    @"directory\n"
-		    @"    -E  --encoding    The encoding used by the archive "
-		    "(only tar and lha files)\n"
-		    @"    -f  --force       Force / overwrite files\n"
-		    @"    -h  --help        Show this help\n"
-		    @"    -l  --list        List all files in the archive\n"
-		    @"    -n  --no-clobber  Never overwrite files\n"
-		    @"    -p  --print       Print one or more files from the "
+		    @"    -E  --encoding=         The encoding used by the "
 		    @"archive\n"
-		    @"    -q  --quiet       Quiet mode (no output, except "
-		    @"errors)\n"
-		    @"    -t  --type        Archive type (gz, lha, tar, tgz, "
-		    @"zip)\n"
-		    @"    -v  --verbose     Verbose output for file list\n"
-		    @"    -x  --extract     Extract files")];
+		    @"                            (only tar, lha and zoo files)"
+		    @"\n"
+		    @"    -f  --force             Force / overwrite files\n"
+		    @"    -h  --help              Show this help\n"
+		    @"        --iri               Use an IRI to access the "
+		    @"archive\n"
+		    @"    -l  --list              List all files in the archive"
+		    @"\n"
+		    @"    -n  --no-clobber        Never overwrite files\n"
+		    @"    -p  --print             Print one or more files from "
+		    @"the archive\n"
+		    @"    -q  --quiet             Quiet mode (no output, "
+		    @"except errors)\n"
+		    @"    -t  --type=             Archive type (gz, lha, tar, "
+		    @"tgz, zip, zoo)\n"
+		    @"    -v  --verbose           Verbose output for file list"
+		    @"\n"
+		    @"    -x  --extract           Extract files")];
 	}
 
 	[OFApplication terminateWithStatus: status];
@@ -100,6 +115,18 @@ mutuallyExclusiveError(OFUnichar shortOption1, OFString *longOption1,
 	[OFApplication terminateWithStatus: 1];
 }
 
+static OFIRI *
+argumentToIRI(OFString *argument, bool isIRI)
+{
+	if (isIRI)
+		return [OFIRI IRIWithString: argument];
+
+	if ([argument isEqual: @"-"])
+		return nil;
+
+	return [OFIRI fileIRIWithPath: argument];
+}
+
 static void
 mutuallyExclusiveError5(OFUnichar shortOption1, OFString *longOption1,
     OFUnichar shortOption2, OFString *longOption2,
@@ -146,7 +173,8 @@ writingNotSupported(OFString *type)
 }
 
 static void
-addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
+addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files,
+    OFString *archiveComment)
 {
 	OFMutableArray *expandedFiles =
 	    [OFMutableArray arrayWithCapacity: files.count];
@@ -163,20 +191,29 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 			[expandedFiles addObject: file];
 	}
 
-	[archive addFiles: expandedFiles];
+	if (expandedFiles.count < 1) {
+		[OFStdErr writeLine: OF_LOCALIZED(@"add_no_file_specified",
+		    @"Need one or more files to add!")];
+		[OFApplication terminateWithStatus: 1];
+	}
+
+	[archive addFiles: expandedFiles archiveComment: archiveComment];
 }
 
 @implementation OFArc
 - (void)applicationDidFinishLaunching: (OFNotification *)notification
 {
-	OFString *outputDir, *encodingString, *type;
+	OFString *archiveComment, *outputDir, *encodingString, *type;
+	bool isIRI;
 	const OFOptionsParserOption options[] = {
 		{ 'a', @"append", 0, NULL, NULL },
+		{ 0,   @"archive-comment", 1, NULL, &archiveComment },
 		{ 'c', @"create", 0, NULL, NULL },
 		{ 'C', @"directory", 1, NULL, &outputDir },
 		{ 'E', @"encoding", 1, NULL, &encodingString },
 		{ 'f', @"force", 0, NULL, NULL },
 		{ 'h', @"help", 0, NULL, NULL },
+		{ 0,   @"iri", 0, &isIRI, NULL },
 		{ 'l', @"list", 0, NULL, NULL },
 		{ 'n', @"no-clobber", 0, NULL, NULL },
 		{ 'p', @"print", 0, NULL, NULL },
@@ -190,6 +227,7 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 	OFStringEncoding encoding = OFStringEncodingAutodetect;
 	OFOptionsParser *optionsParser;
 	OFArray OF_GENERIC(OFString *) *remainingArguments, *files;
+	OFIRI *IRI;
 	id <Archive> archive;
 
 #ifdef OF_HAVE_SANDBOX
@@ -338,12 +376,13 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		if (remainingArguments.count < 1)
 			help(OFStdErr, false, 1);
 
+		IRI = argumentToIRI(remainingArguments.firstObject, isIRI);
 		files = [remainingArguments objectsInRange:
 		    OFMakeRange(1, remainingArguments.count - 1)];
 
 #ifdef OF_HAVE_SANDBOX
-		if (![remainingArguments.firstObject isEqual: @"-"])
-			[sandbox unveilPath: remainingArguments.firstObject
+		if ([IRI.scheme isEqual: @"file"])
+			[sandbox unveilPath: IRI.fileSystemRepresentation
 				permissions: (mode == 'a' ? @"rwc" : @"wc")];
 
 		for (OFString *path in files)
@@ -353,32 +392,32 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		[OFApplication of_activateSandbox: sandbox];
 #endif
 
-		archive = [self
-		    openArchiveWithPath: remainingArguments.firstObject
-				   type: type
-				   mode: mode
-			       encoding: encoding];
+		archive = [self openArchiveWithIRI: IRI
+					      type: type
+					      mode: mode
+					  encoding: encoding];
 
-		addFiles(archive, files);
+		addFiles(archive, files, archiveComment);
 		break;
 	case 'l':
 		if (remainingArguments.count != 1)
 			help(OFStdErr, false, 1);
 
+		IRI = argumentToIRI(remainingArguments.firstObject, isIRI);
+
 #ifdef OF_HAVE_SANDBOX
-		if (![remainingArguments.firstObject isEqual: @"-"])
-			[sandbox unveilPath: remainingArguments.firstObject
+		if ([IRI.scheme isEqual: @"file"])
+			[sandbox unveilPath: IRI.fileSystemRepresentation
 				permissions: @"r"];
 
 		sandbox.allowsUnveil = false;
 		[OFApplication of_activateSandbox: sandbox];
 #endif
 
-		archive = [self
-		    openArchiveWithPath: remainingArguments.firstObject
-				   type: type
-				   mode: mode
-			       encoding: encoding];
+		archive = [self openArchiveWithIRI: IRI
+					      type: type
+					      mode: mode
+					  encoding: encoding];
 
 		[archive listFiles];
 		break;
@@ -386,23 +425,23 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		if (remainingArguments.count < 1)
 			help(OFStdErr, false, 1);
 
+		IRI = argumentToIRI(remainingArguments.firstObject, isIRI);
+		files = [remainingArguments objectsInRange:
+		    OFMakeRange(1, remainingArguments.count - 1)];
+
 #ifdef OF_HAVE_SANDBOX
-		if (![remainingArguments.firstObject isEqual: @"-"])
-			[sandbox unveilPath: remainingArguments.firstObject
+		if ([IRI.scheme isEqual: @"file"])
+			[sandbox unveilPath: IRI.fileSystemRepresentation
 				permissions: @"r"];
 
 		sandbox.allowsUnveil = false;
 		[OFApplication of_activateSandbox: sandbox];
 #endif
 
-		files = [remainingArguments objectsInRange:
-		    OFMakeRange(1, remainingArguments.count - 1)];
-
-		archive = [self
-		    openArchiveWithPath: remainingArguments.firstObject
-				   type: type
-				   mode: mode
-			       encoding: encoding];
+		archive = [self openArchiveWithIRI: IRI
+					      type: type
+					      mode: mode
+					  encoding: encoding];
 
 		[archive printFiles: files];
 		break;
@@ -410,12 +449,13 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		if (remainingArguments.count < 1)
 			help(OFStdErr, false, 1);
 
+		IRI = argumentToIRI(remainingArguments.firstObject, isIRI);
 		files = [remainingArguments objectsInRange:
 		    OFMakeRange(1, remainingArguments.count - 1)];
 
 #ifdef OF_HAVE_SANDBOX
-		if (![remainingArguments.firstObject isEqual: @"-"])
-			[sandbox unveilPath: remainingArguments.firstObject
+		if ([IRI.scheme isEqual: @"file"])
+			[sandbox unveilPath: IRI.fileSystemRepresentation
 				permissions: @"r"];
 
 		if (files.count > 0)
@@ -436,11 +476,23 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		[OFApplication of_activateSandbox: sandbox];
 #endif
 
-		archive = [self
-		    openArchiveWithPath: remainingArguments.firstObject
-				   type: type
-				   mode: mode
-			       encoding: encoding];
+		archive = [self openArchiveWithIRI: IRI
+					      type: type
+					      mode: mode
+					  encoding: encoding];
+
+#ifdef OF_MACOS
+		@try {
+			OFString *attributeName = @"com.apple.quarantine";
+
+			_quarantine = [[[OFFileManager defaultManager]
+			    extendedAttributeDataForName: attributeName
+					     ofItemAtIRI: IRI] retain];
+		} @catch (OFGetItemAttributesFailedException *e) {
+			if (e.errNo != /*ENOATTR*/ 93)
+				@throw e;
+		}
+#endif
 
 		if (outputDir != nil) {
 			OFFileManager *fileManager =
@@ -456,26 +508,20 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		@try {
 			[archive extractFiles: files];
 		} @catch (OFCreateDirectoryFailedException *e) {
-			OFString *error = [OFString
-			    stringWithCString: strerror(e.errNo)
-				     encoding: [OFLocale encoding]];
 			[OFStdErr writeString: @"\r"];
 			[OFStdErr writeLine: OF_LOCALIZED(
 			    @"failed_to_create_directory",
 			    @"Failed to create directory %[dir]: %[error]",
 			    @"dir", e.IRI.fileSystemRepresentation,
-			    @"error", error)];
+			    @"error", OFStrError(e.errNo))];
 			_exitStatus = 1;
 		} @catch (OFOpenItemFailedException *e) {
-			OFString *error = [OFString
-			    stringWithCString: strerror(e.errNo)
-				     encoding: [OFLocale encoding]];
 			[OFStdErr writeString: @"\r"];
 			[OFStdErr writeLine: OF_LOCALIZED(
 			    @"failed_to_open_file",
 			    @"Failed to open file %[file]: %[error]",
 			    @"file", e.path,
-			    @"error", error)];
+			    @"error", OFStrError(e.errNo))];
 			_exitStatus = 1;
 		}
 
@@ -488,30 +534,24 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 	[OFApplication terminateWithStatus: _exitStatus];
 }
 
-- (id <Archive>)openArchiveWithPath: (OFString *)path_
-			       type: (OFString *)type
-			       mode: (char)mode
-			   encoding: (OFStringEncoding)encoding
+- (id <Archive>)openArchiveWithIRI: (OFIRI *)IRI
+			      type: (OFString *)type
+			      mode: (char)mode
+			  encoding: (OFStringEncoding)encoding
 {
 	/* To make clang-analyzer happy about assigning nil to path later. */
-	OFString *path = path_;
 	OFString *modeString, *fileModeString;
 	OFStream *file = nil;
 	id <Archive> archive = nil;
 
-	[_archivePath release];
-	_archivePath = [path copy];
-
-	if (path == nil)
-		return nil;
-
 	switch (mode) {
 	case 'a':
 		modeString = @"a";
 		fileModeString = @"r+";
 		break;
 	case 'c':
-		modeString = fileModeString = @"w";
+		modeString = @"w";
+		fileModeString = @"w+";
 		break;
 	case 'l':
 	case 'p':
@@ -522,7 +562,7 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		@throw [OFInvalidArgumentException exception];
 	}
 
-	if ([path isEqual: @"-"]) {
+	if (IRI == nil) {
 		switch (mode) {
 		case 'a':
 		case 'c':
@@ -536,69 +576,77 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 		default:
 			@throw [OFInvalidArgumentException exception];
 		}
-
-		path = nil;
 	} else {
 		@try {
-			file = [OFFile fileWithPath: path mode: fileModeString];
+			file = [OFIRIHandler openItemAtIRI: IRI
+						      mode: fileModeString];
 		} @catch (OFOpenItemFailedException *e) {
-			OFString *error = [OFString
-			    stringWithCString: strerror(e.errNo)
-				     encoding: [OFLocale encoding]];
 			[OFStdErr writeString: @"\r"];
 			[OFStdErr writeLine: OF_LOCALIZED(
 			    @"failed_to_open_file",
 			    @"Failed to open file %[file]: %[error]",
-			    @"file", e.path,
-			    @"error", error)];
+			    @"file", e.IRI.string,
+			    @"error", OFStrError(e.errNo))];
 			[OFApplication terminateWithStatus: 1];
 		}
 	}
 
 	if (type == nil || [type isEqual: @"auto"]) {
+		OFString *lowercasePath = IRI.path.lowercaseString;
+
 		/* This one has to be first for obvious reasons */
-		if ([path hasSuffix: @".tar.gz"] || [path hasSuffix: @".tgz"] ||
-		    [path hasSuffix: @".TAR.GZ"] || [path hasSuffix: @".TGZ"])
+		if ([lowercasePath hasSuffix: @".tar.gz"] ||
+		    [lowercasePath hasSuffix: @".tgz"])
 			type = @"tgz";
-		else if ([path hasSuffix: @".gz"] || [path hasSuffix: @".GZ"])
+		else if ([lowercasePath hasSuffix: @".gz"])
 			type = @"gz";
-		else if ([path hasSuffix: @".lha"] || [path hasSuffix: @".lzh"])
+		else if ([lowercasePath hasSuffix: @".lha"] ||
+		    [lowercasePath hasSuffix: @".lzh"] ||
+		    [lowercasePath hasSuffix: @".lzs"] ||
+		    [lowercasePath hasSuffix: @".pma"])
 			type = @"lha";
-		else if ([path hasSuffix: @".tar"] || [path hasSuffix: @".TAR"])
+		else if ([lowercasePath hasSuffix: @".tar"])
 			type = @"tar";
+		else if ([lowercasePath hasSuffix: @".zoo"])
+			type = @"zoo";
 		else
 			type = @"zip";
 	}
 
 	@try {
 		if ([type isEqual: @"gz"])
-			archive = [GZIPArchive archiveWithPath: path
-							stream: file
-							  mode: modeString
-						      encoding: encoding];
-		else if ([type isEqual: @"lha"])
-			 archive = [LHAArchive archiveWithPath: path
-							stream: file
-							  mode: modeString
-						      encoding: encoding];
-		else if ([type isEqual: @"tar"])
-			archive = [TarArchive archiveWithPath: path
+			archive = [GZIPArchive archiveWithIRI: IRI
 						       stream: file
 							 mode: modeString
 						     encoding: encoding];
+		else if ([type isEqual: @"lha"])
+			 archive = [LHAArchive archiveWithIRI: IRI
+						       stream: file
+							 mode: modeString
+						      encoding: encoding];
+		else if ([type isEqual: @"tar"])
+			archive = [TarArchive archiveWithIRI: IRI
+						      stream: file
+							mode: modeString
+						    encoding: encoding];
 		else if ([type isEqual: @"tgz"]) {
 			OFStream *GZIPStream = [OFGZIPStream
 			    streamWithStream: file
 					mode: modeString];
-			archive = [TarArchive archiveWithPath: path
-						       stream: GZIPStream
-							 mode: modeString
-						     encoding: encoding];
+			archive = [TarArchive archiveWithIRI: IRI
+						      stream: GZIPStream
+							mode: modeString
+						    encoding: encoding];
 		} else if ([type isEqual: @"zip"])
-			archive = [ZIPArchive archiveWithPath: path
-						       stream: file
-							 mode: modeString
-						     encoding: encoding];
+			archive = [ZIPArchive archiveWithIRI: IRI
+						      stream: file
+							mode: modeString
+						    encoding: encoding];
+		else if ([type isEqual: @"zoo"])
+			archive = [ZooArchive archiveWithIRI: IRI
+						      stream: file
+							mode: modeString
+						    encoding: encoding];
 		else {
 			[OFStdErr writeLine: OF_LOCALIZED(
 			    @"unknown_archive_type",
@@ -615,33 +663,27 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 
 		@throw e;
 	} @catch (OFReadFailedException *e) {
-		OFString *error = [OFString
-		    stringWithCString: strerror(e.errNo)
-			     encoding: [OFLocale encoding]];
 		[OFStdErr writeLine: OF_LOCALIZED(@"failed_to_read_file",
 		    @"Failed to read file %[file]: %[error]",
-		    @"file", path,
-		    @"error", error)];
+		    @"file", IRI.string,
+		    @"error", OFStrError(e.errNo))];
 		goto error;
 	} @catch (OFSeekFailedException *e) {
-		OFString *error = [OFString
-		    stringWithCString: strerror(e.errNo)
-			     encoding: [OFLocale encoding]];
 		[OFStdErr writeLine: OF_LOCALIZED(@"failed_to_seek_in_file",
 		    @"Failed to seek in file %[file]: %[error]",
-		    @"file", path,
-		    @"error", error)];
+		    @"file", IRI.string,
+		    @"error", OFStrError(e.errNo))];
 		goto error;
 	} @catch (OFInvalidFormatException *e) {
 		[OFStdErr writeLine: OF_LOCALIZED(
 		    @"file_is_not_a_valid_archive",
 		    @"File %[file] is not a valid archive!",
-		    @"file", path)];
+		    @"file", IRI.string)];
 		goto error;
 	}
 
-	if ((mode == 'a' || mode == 'c') &&
-	    ![archive respondsToSelector: @selector(addFiles:)]) {
+	if ((mode == 'a' || mode == 'c') && ![archive respondsToSelector:
+	    @selector(addFiles:archiveComment:)]) {
 		writingNotSupported(type);
 		goto error;
 	}
@@ -649,8 +691,8 @@ addFiles(id <Archive> archive, OFArray OF_GENERIC(OFString *) *files)
 	return archive;
 
 error:
-	if (mode == 'c' && path != nil)
-		[[OFFileManager defaultManager] removeItemAtPath: path];
+	if (mode == 'c' && IRI != nil)
+		[[OFFileManager defaultManager] removeItemAtIRI: IRI];
 
 	[OFApplication terminateWithStatus: 1];
 	abort();
@@ -725,28 +767,22 @@ error:
 	@try {
 		length = [input readIntoBuffer: buffer length: bufferSize];
 	} @catch (OFReadFailedException *e) {
-		OFString *error = [OFString
-		    stringWithCString: strerror(e.errNo)
-			     encoding: [OFLocale encoding]];
 		[OFStdOut writeString: @"\r"];
 		[OFStdErr writeLine: OF_LOCALIZED(@"failed_to_read_file",
 		    @"Failed to read file %[file]: %[error]",
 		    @"file", fileName,
-		    @"error", error)];
+		    @"error", OFStrError(e.errNo))];
 		return -1;
 	}
 
 	@try {
 		[output writeBuffer: buffer length: length];
 	} @catch (OFWriteFailedException *e) {
-		OFString *error = [OFString
-		    stringWithCString: strerror(e.errNo)
-			     encoding: [OFLocale encoding]];
 		[OFStdOut writeString: @"\r"];
 		[OFStdErr writeLine: OF_LOCALIZED(@"failed_to_write_file",
 		    @"Failed to write file %[file]: %[error]",
 		    @"file", fileName,
-		    @"error", error)];
+		    @"error", OFStrError(e.errNo))];
 		return -1;
 	}
 
@@ -798,4 +834,15 @@ error:
 
 	return [path autorelease];
 }
+
+- (void)quarantineFile: (OFString *)path
+{
+#ifdef OF_MACOS
+	if (_quarantine != nil)
+		[[OFFileManager defaultManager]
+		    setExtendedAttributeData: _quarantine
+				     forName: @"com.apple.quarantine"
+				ofItemAtPath: path];
+#endif
+}
 @end
